import { reactive, ref, toRefs, unref, watch } from 'vue'
import type { Ref } from 'vue'
import { debounce } from 'lodash-es'
import { createFakeDom } from '@/utils/common'

export type Point = [number, number]

export type RGBA = number[]

export interface GetPixPointParams {
  width: number
  height: number
}

/**
 * @description 相似度阀值
 */
export const DISTANCE = 45

/**
 * @description 黑色RGBA
 */
export const BLACK = [0, 0, 0, 255]

/**
 * @description 白色RGBA
 */
export const WHITE = [255, 255, 255, 255]

/**
 * @description 背景色取色点
 */
export const COLOR_POINT: Point[] = [
  [0, 0],
  [0.25, 0],
  [0.5, 0],
  [0.75, 0],
  [1, 0],
  [0, 0.25],
  [1, 0.25],
  [0, 0.5],
  [1, 0.5],
  [0, 0.75],
  [1, 0.75],
  [0, 1],
  [0.25, 1],
  [0.5, 1],
  [0.75, 1],
  [1, 1],
]

/**
 * @description 根据取色点获取图片像素点
 * @params { Object } width canvas宽度
 * @params { Object } height canvas高度
 */

export const getPixelPoint = ({ width, height }: GetPixPointParams, colorPoint: Point[] = COLOR_POINT) => {
  return colorPoint.map(([x, y]) => {
    const index = Math.max(Math.ceil(width * x), 0) + Math.max(Math.ceil(height * y) - 1, 0) * width
    return Math.max(index + Math.ceil(x) * -1, 0) * 4
  })
}

/**
 * @description 计算像素点相似度
 */
export const calcPixSim = (rgba: RGBA, base: RGBA = WHITE) => {
  const [r, g, b, a] = rgba
  const [r1, g1, b1, a1] = base
  return Math.sqrt((r - r1) ** 2 + (g - g1) ** 2 + (b - b1) ** 2 + (a - a1) ** 2)
}

/**
 * @description 判断两个颜色是否相似
 */
export const isSimilar = (color1: RGBA, color2: RGBA, distance: number = DISTANCE) => {
  return calcPixSim(color1, color2) <= distance
}

/**
 * @description 根据imageData起始下标获取rgba
 */
export const getColorFromIndex = (index: number, data: Uint8ClampedArray) => {
  return [0, 1, 2, 3].map((_) => {
    return data[index + _]
  })
}

/**
 * @description 根据imageData起始下标设置rgba
 */
export const setColorFromIndex = (index: number, data: Uint8ClampedArray, color: RGBA) => {
  return [0, 1, 2, 3].forEach((_) => {
    data[index + _] = color[_]
  })
}

/**
 * @description 获取各取色点中出现次数最多的颜色
 */
export const getMainColor = (colors: RGBA[]) => {
  let max = 1
  let index = 0
  let num = 0
  const total = colors
    .map((_) => calcPixSim(_))
    .reduce((r, n, i) => {
      r[n] ??= 0
      r[n]++
      if (r[n] > max) {
        max = r[n]
        num = n
        index = i
      }
      return r
    }, {} as Record<number, number>)
  return {
    max,
    index,
    num,
    total,
    mainColor: colors[index],
  }
}

/**
 * @description 从字符串中提取颜色值
 * @example "rgba(121,121,121,1)"
 */

export const convertColor = (color: string) => {
  const [r, g, b, a] = color.replace(/rgba\((.*)\)/g, '$1').split(',') as unknown as RGBA
  return [r, g, b, a * 255]
}

export interface DrawCanvasOption {
  image: HTMLImageElement | null
  canvas?: HTMLCanvasElement | null
  useNaturalSize?: boolean
  basePoint?: Point[]
  bgColor?: RGBA
  distance?: number
  markColor?: RGBA
}

export interface DrawCanvasResult {
  imageData: ImageData | null
  pixelPoints: number[]
  pixelsInfo: RGBA[]
  dataUrl: string
}

export interface SetImgBgOption extends DrawCanvasOption {
  basePoint?: Point[]
  useNaturalSize?: boolean
  bgColor?: RGBA
  distance?: number
  markColor?: RGBA
  debounceDelay?: number
}

export const drawCanvas = async ({
  image,
  canvas,
  useNaturalSize,
  basePoint,
  bgColor,
  distance,
  markColor,
}: DrawCanvasOption) => {
  const result: DrawCanvasResult = {
    imageData: null,
    pixelsInfo: [],
    pixelPoints: [],
    dataUrl: '',
  }
  try {
    await new Promise((resolve) => setTimeout(resolve, 20))
    if (!image) {
      console.warn(`return for img define ${image}`)
      return result
    }

    const c = canvas || createFakeDom('canvas')
    const ctx = c.getContext('2d')

    if (!ctx) {
      console.warn(`return for canvas context define ${ctx}`)
      return result
    }

    const width: number = (c.width = useNaturalSize ? image.naturalWidth : image.width)
    const height: number = (c.height = useNaturalSize ? image.naturalHeight : image.height)

    if (width <= 0 || height <= 0) {
      console.warn(`return for img size define width: ${width}, height: ${height}`)
      return result
    }

    ctx.drawImage(image, 0, 0, width, height)

    const imageData = ctx.getImageData(0, 0, width, height)

    const pixelPoints = getPixelPoint({ width: width, height: height }, basePoint)

    const pixelsInfo = pixelPoints.map((_) => getColorFromIndex(_, imageData.data))
    const { mainColor } = getMainColor(pixelsInfo)

    if (bgColor) {
      if (isSimilar(bgColor, mainColor, distance)) {
        console.log('原样输出', mainColor, bgColor)
        return Object.assign(result, {
          imageData,
          pixelsInfo,
          pixelPoints,
          dataUrl: c.toDataURL(),
        })
      }

      for (let index = 0; index < imageData.data.length; index += 4) {
        if (markColor && pixelPoints.includes(index)) {
          setColorFromIndex(index, imageData.data, markColor)
        }
        if (isSimilar(getColorFromIndex(index, imageData.data), BLACK, distance)) {
          setColorFromIndex(index, imageData.data, BLACK)
        }
        if (isSimilar(getColorFromIndex(index, imageData.data), mainColor, distance)) {
          setColorFromIndex(index, imageData.data, bgColor)
        }
      }

      ctx.putImageData(imageData, 0, 0)
    }
    return Object.assign(result, {
      ImageData,
      pixelsInfo,
      pixelPoints,
      dataUrl: c.toDataURL(),
    })
  } catch (error) {
    console.error(error)
    return result
  }
}

export const useSetImgBg = (option: Ref<SetImgBgOption>) => {
  const drawing = ref<boolean>(false)
  const imageData = ref<ImageData | null>(null)
  const pixelPoints = ref<number[]>([])
  const dataUrl = ref('')
  const width = ref<number>(0)
  const height = ref<number>(0)

  const update = debounce(async () => {
    drawing.value = true
    const result = await drawCanvas(unref(option))
    dataUrl.value = result.dataUrl
    imageData.value = result.imageData
    pixelPoints.value = result.pixelPoints
    drawing.value = false
  }, option.value.debounceDelay || 300)

  watch(option, update, { deep: true })
  return toRefs(
    reactive({
      drawing,
      width,
      height,
      imageData,
      pixelPoints,
      dataUrl,
      update,
    })
  )
}
